# pylint: skip-file

from aioasuswrt.helpers import convert_size


def test_convert_size() -> None:
    assert "0 B" == convert_size(0)
    assert "1.0 B" == convert_size(1)
    assert "1.0 KB" == convert_size(1024)
    assert "1.0 MB" == convert_size(1024 * 1024)
    assert "1.0 GB" == convert_size(1024 * 1024 * 1024)



"""Mappings of structures and typing."""

from enum import StrEnum
from re import Pattern
from re import compile as re_compile
from typing import Callable, NamedTuple, TypeAlias, TypedDict

_VPN_COUNT: int = 5  # Max allowed VPN setups


InterfaceJson: TypeAlias = dict[
    str, dict[str, dict[str, dict[str, str | int]]]
]


class AsyncSSHConnectKwargs(NamedTuple):
    """Kwargs mapping for the asyncssh.connect method."""

    username: str
    client_keys: list[str] | None
    port: int
    password: str | None
    passphrase: str | None
    known_host: list[str] | None = None
    server_host_key_algs: list[str] = [
        "ssh-rsa",
        "rsa-sha2-256",
        "rsa-sha2-512",
        "ecdsa-sha2-nistp256",
        "ecdsa-sha2-nistp384",
        "ecdsa-sha2-nistp521",
        "ssh-ed25519",
        "ssh-ed448",
    ]


class Interface(TypedDict):
    """
    Interface representation.

    Attributes:
        id (str | None): id of the interface (for example eth0)
        name (str | None): Name of the interface (for example 5g)
        mac (str | None): MAC address for the interface
    """

    id: str | None
    name: str | None
    mac: str | None


class DeviceData(TypedDict):
    """
    Device status representation.

    Attributes:
        ip (str |None): The IP of the device
        name (str | None): The hostname of the device
        status (str | None): The status of the device
        rssi (int | None): Signal strength,
            in a mesh systems this is to closest node.
    """

    ip: str | None
    name: str | None
    status: str | None
    rssi: int | None


class Device(NamedTuple):
    """
    Representation of a connected device.

    Attributes:
        mac (str): The MAC adrress for the device
        device_data (DeviceData): The current status
        interface (Interface): Information about the
            Interface device is connected to
    """

    mac: str
    device_data: DeviceData
    interface: Interface


class TransferRates(NamedTuple):
    """
    Representation of transfer rates.

    Attributes:
        rx (int): Received bytes
        tx (int): Transferred bytes
    """

    rx: int = 0
    tx: int = 0


class ConnectionType(StrEnum):
    """Connection type definition."""

    SSH = "SSH"
    TELNET = "TELNET"


class AuthConfig(TypedDict):
    """
    Authentication configuration

    There are multiple ways to connect to the router,
    we recomend using ssh_key with a passphrase if possible

    Attributes:
        username (str | None): The username to use
        password (str | None): The password to use
            required if no ssh_key is set
        connection_type (ConnectionType | None):
            Defaults to ConnectionType.SSH
        ssh_key (str |None): An optional ssh_key
        passphrase (str |None): An optional passphrase, used for ssh_key
        port (int | None): Defaults to 22 for ssh and 110 for telnet
    """

    username: str | None
    password: str | None
    connection_type: ConnectionType | None
    ssh_key: str | None
    passphrase: str | None
    port: int


class Mode(StrEnum):
    """Router modes definition."""

    ROUTER = "router"
    AP = "ap"


class Settings(NamedTuple):
    """
    Settings for communicating with asuswrt router.

    Args:
        require_ip (bool | None): Defaults to False
            if set to True we will not fetch any device we cannot map ip for
        mode (Mode | None): Defaults to Mode.ROUTER
        dnsmasq (str | None): Defaults to "/var/lib/misc"
            Directory where dnsmasq.leases can be found
        wan_interface (str | None): Defaults to eth0
            The name of the WAN interface connection
            used to get external IP and transfer rates
    """

    require_ip: bool = False
    mode: Mode | None = Mode.ROUTER
    dnsmasq: str = "/var/lib/misc"
    wan_interface: str = "eth0"


class TempCommand(NamedTuple):
    """
    Representation of a temperature command.

    Attributes:
        cli_command (str): The actual cli-command to run
            example "cat /proc/version"
        result_location (int): We .split(" ") the result of the command
            this tells which index of the resulting list to use
        eval_function (Callable): Method we run on the retrieved value
    """

    cli_command: str
    result_location: int
    eval_function: Callable[..., float]


class Regex(NamedTuple):
    """Regex Mapped to a key."""

    VPN_LIST: Pattern[str] = re_compile(
        (
            r"(?P<description>.+?)>"
            r"(?P<type>.+?)>"
            r"(?P<id>.+?)>"
            r"(?P<username>.*?)>"
            r"(?P<password>.*?)(?:<|$)"
        )
    )
    LEASES: Pattern[str] = re_compile(
        (
            r"\w+\s"
            r"(?P<mac>(([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})))\s"
            r"(?P<ip>([0-9]{1,3}[\.]){3}[0-9]{1,3})\s"
            r"(?P<host>([^\s]+))"
        )
    )

    WL: Pattern[str] = re_compile(
        (
            r"\w+\s"
            r"(?P<mac>(([0-9A-F]{2}[:-]){5}([0-9A-F]{2})))"
        )
    )
    NVRAM: str = "^{}=([\\w.\\-/: <>]+)"
    IP_NEIGH: Pattern[str] = re_compile(
        (
            r"(?P<ip>([0-9]{1,3}[\.]){3}[0-9]{1,3}|"
            r"([0-9a-fA-F]{1,4}:){1,7}"
            r"[0-9a-fA-F]{0,4}(:[0-9a-fA-F]{1,4}){1,7})\s"
            r"\w+\s"
            r"\w+.+\s"
            r"(\w+\s(?P<mac>(([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2}))))?\s"
            r"\s?(router)?"
            r"\s?(nud)?"
            r"(?P<status>(\w+))"
        )
    )
    ARP: Pattern[str] = re_compile(
        (
            r".+\s"
            r"\((?P<ip>([0-9]{1,3}[\.]){3}[0-9]{1,3})\)\s"
            r".+\s"
            r"(?P<mac>(([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})))"
            r".+\s"
            r"(?P<interface>([\w.-]+))"
        )
    )


compiled_regex = Regex()


class Command(StrEnum):
    """List of commands that we run."""

    GET_PID_OF = "pidof {name}"
    NETDEV = "cat /proc/net/dev"
    RX = "cat /sys/class/net/{}/statistics/rx_bytes"
    TX = "cat /sys/class/net/{}/statistics/tx_bytes"

    MEMINFO = "cat /proc/meminfo"
    LOADAVG = "cat /proc/loadavg"
    ADDHOST = (
        'cat /etc/hosts | grep -q "{ipaddress} {hostname}" || '
        '(echo "{ipaddress} {hostname}" >> /etc/hosts && '
        "kill -HUP `cat /var/run/dnsmasq.pid`)"
    )
    LEASES = "cat {}/dnsmasq.leases"
    WL = (
        "for dev in `nvram get wl1_vifs && nvram get wl0_vifs && "
        "nvram get wl_ifnames`; do "
        "if type wlanconfig > /dev/null; then "
        "wlanconfig $dev list | awk 'FNR > 1 {print substr($1, 0, 18)}';"
        " else wl -i $dev assoclist; fi; done"
    )
    CLIENTLIST = "cat /tmp/clientlist.json"
    NVRAM = "nvram show"
    IP_NEIGH = "ip neigh"
    ARP = "arp -n"
    VPN_START = "service start_vpnclient{id}"
    VPN_STOP = "service stop_vpnclient{id}"


_NVRAM_VPN_LIST: list[str] = ["vpnc_clientlist"]
_NVRAM_VPN_LIST.extend([f"vpn_client{i + 1}_state" for i in range(_VPN_COUNT)])


class Nvram(NamedTuple):
    """Property mapping for the nvram info command."""

    DHCP: list[str] = [
        "dhcp_dns1_x",
        "dhcp_dns2_x",
        "dhcp_enable_x",
        "dhcp_start",
        "dhcp_end",
        "dhcp_lease",
    ]
    MODEL: list[str] = ["model"]
    QOS: list[str] = [
        "qos_ack",
        "qos_atm",
        "qos_burst0",
        "qos_burst1",
        "qos_default",
        "qos_enable",
        "qos_fin",
        "qos_ibw",
        "qos_ibw1",
        "qos_icmp",
        "qos_irates",
        "qos_method",
        "qos_obw",
        "qos_obw1",
        "qos_orules",
        "qos_overhead",
        "qos_reset",
        "qos_rst",
        "qos_sched",
        "qos_sticky",
        "qos_syn",
        "qos_type",
    ]
    REBOOT: list[str] = [
        "reboot_schedule",
        "reboot_schedule_enable",
        "reboot_time",
    ]
    WLAN: list[str] = [
        "wan_dns",
        "wan_domain",
        "wan_enable",
        "wan_expires",
        "wan_gateway",
        "wan_ipaddr",
        "wan_lease",
        "wan_mtu",
        "wan_realip_ip",
        "wan_realip_state",
    ]
    GUEST_2G_1: list[str] = [
        "wl0.1_bss_enabled",
        "wl0.1_lanaccess",
        "wl0.1_ssid",
        "wl0.1_wpa_psk",
    ]
    GUEST_2G_2: list[str] = [
        "wl0.2_bss_enabled",
        "wl0.2_lanaccess",
        "wl0.2_ssid",
        "wl0.2_wpa_psk",
    ]
    GUEST_2G_3: list[str] = [
        "wl0.3_bss_enabled",
        "wl0.3_lanaccess",
        "wl0.3_ssid",
        "wl0.3_wpa_psk",
    ]
    WIFI_2G: list[str] = [
        "wl0_bss_enabled",
        "wl0_chanspec",
        "wl0_ssid",
        "wl0_wpa_psk",
    ]
    GUEST_5G_1: list[str] = [
        "wl1.1_bss_enabled",
        "wl1.1_lanaccess",
        "wl1.1_ssid",
        "wl1.1_wpa_psk",
    ]
    GUEST_5G_2: list[str] = [
        "wl1.2_bss_enabled",
        "wl1.2_lanaccess",
        "wl1.2_ssid",
        "wl1.2_wpa_psk",
    ]
    GUEST_5G_3: list[str] = [
        "wl1.3_bss_enabled",
        "wl1.3_lanaccess",
        "wl1.3_ssid",
        "wl1.3_wpa_psk",
    ]
    WIFI_5G: list[str] = [
        "wl1_bss_enabled",
        "wl1_chanspec",
        "wl1_ssid",
        "wl1_wpa_psk",
    ]
    FIRMWARE: list[str] = [
        "buildinfo",
        "buildno",
        "buildno_org",
        "firmver",
        "firmware_check",
        "firmware_check_enable",
        "firmware_path",
        "firmware_server",
        "webs_last_info",
        "webs_notif_flag",
        "webs_state_REQinfo",
        "webs_state_error",
        "webs_state_flag",
        "webs_state_odm",
        "webs_state_update",
        "webs_state_upgrade",
        "webs_state_url",
    ]
    LABEL_MAC: list[str] = ["label_mac"]
    VPN: list[str] = _NVRAM_VPN_LIST



"""AioAsusWrt helpers."""

from math import floor, log
from math import pow as mpow


def convert_size(size: int) -> str:
    """
    Convert bytes to a readable string with MB/GB and so on added as a suffix.

    Args:
        size (float): Size given in number of bytes

    """
    if size == 0:
        return "0 B"
    name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    index = int(floor(log(size, 1024)))
    power = mpow(1024, index)
    return f"{round(size / power, 2)} {name[index]}"



"""Module for Asuswrt class."""

from collections.abc import Iterable
from json import loads
from logging import getLogger
from re import Pattern, findall, finditer, split
from time import time
from typing import final

from .connection import BaseConnection, create_connection
from .constant import NETDEV_FIELDS, TEMP_COMMANDS, VPN_COUNT
from .helpers import convert_size
from .structure import (
    AuthConfig,
    Command,
    Device,
    DeviceData,
    Interface,
    InterfaceJson,
    Mode,
    Nvram,
    Settings,
    TempCommand,
    TransferRates,
)
from .structure import (
    compiled_regex as Regex,
)

_LOGGER = getLogger(__name__)


async def _run_temp_command(
    api: BaseConnection, command: TempCommand
) -> float | None:
    command_result: list[str] = (
        await api.run_command(str(command.cli_command))
    )[0].split(" ")
    result = command_result[int(command.result_location)]
    if result.isnumeric():
        return float(command.eval_function(float(result)))
    return None


async def _parse_lines(
    lines: Iterable[str], regex: Pattern[str]
) -> Iterable[dict[str, str]]:
    """
    Parse the lines using the given regular expression.

    If a line can't be parsed it is logged and skipped in the output.
    We first map all the lines, and after this filter out any None rows.

    Args:
        lines (Iterable[str]): A list of lines to parse
        regex (Pattern[str]): The regex pattern to use on each line
    """

    def _match(line: str) -> dict[str, str] | None:
        """
        Match a single line, will return a None value if no match is made.

        Args:
            line (str): single line to process
        """
        if not line:
            return None
        match = regex.search(line)
        if not match:
            _LOGGER.debug("Could not parse row: %s", line)
            return None
        return match.groupdict()

    return list(filter(None, map(_match, lines)))


def _new_device(mac: str) -> Device:
    """
    Initialize a new device.

    Args:
        mac (str): The MAC address of the device
    """
    return Device(
        mac,
        DeviceData(ip=None, name=None, status=None, rssi=None),
        Interface(id=None, name=None, mac=None),
    )


@final
class AsusWrt:
    """This is the main interface class."""

    def __init__(
        self,
        host: str,
        auth_config: AuthConfig,
        settings: Settings | None = None,
    ) -> None:
        """
        Initiate the AsusWrt class, and the connection interface.

        Args:
            host (str): IP or hostname for the router
            auth_config (AuthConfig): The authentication configuration
            settings (Settings | None): Optional aioasuswrt settings
        """
        self._settings = settings or Settings()
        self._transfer_rates = TransferRates()
        self._last_transfer_rates_check = time()
        self._temps_commands: dict[str, TempCommand] = {}
        self._connection = create_connection(
            host,
            auth_config,
        )

    @property
    def wan_interface(self) -> str:
        """Wan interface property."""
        return str(self._settings.wan_interface)

    async def get_nvram(self, parameter_to_fetch: str) -> dict[str, str]:
        """
        Get nvram value.

        Args:
            parameter_to_fetch (str): The parameter we are targeting to fetch.
        """
        data: dict[str, str] = {}
        target: list[str] | None = getattr(Nvram, parameter_to_fetch, None)
        if not isinstance(target, list):
            return data

        lines = await self._connection.run_command(Command.NVRAM)
        if not lines:
            _LOGGER.warning("No devices found in router")
            return data
        for item in target:
            regex = Regex.NVRAM.format(item)
            for line in lines:
                result = findall(regex, line)
                if result:
                    data[item] = result[0]
                    break
        return data

    async def _get_wl(self, devices: dict[str, Device]) -> dict[str, Device]:
        """
        Add devices fount in wl.

        Args:
            devices (dict[str, Device]): Currently known device list
        """
        _LOGGER.info("get_wl")
        lines = await self._connection.run_command(Command.WL)
        if not lines:
            return devices
        result = await _parse_lines(lines, Regex.WL)

        def _handle(device: dict[str, str]) -> None:
            mac = device["mac"].upper()
            devices[mac] = _new_device(mac)

        _ = list(map(_handle, result))
        _LOGGER.info("There are %s devices found in wl", len(devices))
        return devices

    async def _get_arp(self, devices: dict[str, Device]) -> dict[str, Device]:
        """
        Add devices found in arp.

        Args:
            devices (dict[str, Device]): Currently known device list
        """
        _LOGGER.info("get_arp")
        lines = await self._connection.run_command(Command.ARP)
        if not lines:
            return devices
        result = await _parse_lines(lines, Regex.ARP)

        def _handle(device: dict[str, str]) -> None:
            mac = device["mac"].upper()
            if mac not in devices:
                devices[mac] = _new_device(mac)
            devices[mac].device_data["ip"] = device["ip"]
            devices[mac].interface["id"] = device["interface"]

        _ = list(map(_handle, result))
        _LOGGER.info("There are %s devices found in arp", len(devices))
        return devices

    async def _get_leases(
        self, devices: dict[str, Device]
    ) -> dict[str, Device]:
        """
        Add devices found in leases.

        Args:
            devices (dict[str, Device]): Currently known device list
        """
        _LOGGER.info("get_leases")
        lines: Iterable[str] = await self._connection.run_command(
            Command.LEASES.format(self._settings.dnsmasq)
        )
        if not lines:
            return devices
        lines = filter(lambda line: not line.startswith("duid "), lines)
        result = await _parse_lines(lines, Regex.LEASES)

        def _handle(device: dict[str, str]) -> None:
            mac = device["mac"].upper()
            if mac not in devices:
                devices[mac] = _new_device(mac)
            host = device.get("host")
            if host is not None and host != "*":
                devices[mac].device_data["name"] = host

            devices[mac].device_data["ip"] = device["ip"]

        _ = list(map(_handle, result))
        _LOGGER.info("There are %s devices found after leases", len(devices))
        return devices

    async def _get_neigh(
        self, devices: dict[str, Device]
    ) -> dict[str, Device]:
        """
        Add devices found in neigh.

        Args:
            devices (dict[str, Device]): Currently known device list
        """
        _LOGGER.info("get_neigh")
        lines = await self._connection.run_command(Command.IP_NEIGH)
        if not lines:
            return devices
        result = await _parse_lines(lines, Regex.IP_NEIGH)

        def _handle(device: dict[str, str]) -> None:
            if not device.get("mac"):
                return
            status = device["status"]
            mac = device["mac"].upper()
            if mac not in devices:
                devices[mac] = _new_device(mac)
            devices[mac].device_data["status"] = status
            devices[mac].device_data["ip"] = device.get(
                "ip", devices[mac].device_data["ip"]
            )

        _ = list(map(_handle, result))
        _LOGGER.info("There are %s devices found in neigh", len(devices))
        return devices

    async def _filter_dev_list(
        self, devices: dict[str, Device]
    ) -> dict[str, Device]:
        """
        Filter devices list using 'clientlist.json' file if available.

        Args:
            devices (dict[str, Device]): Currently known device list
        """
        _LOGGER.info("filter_dev_list")
        lines = await self._connection.run_command(Command.CLIENTLIST)
        if not lines:
            return devices

        def _parse_client_json(data: str) -> None:
            device_list = InterfaceJson(loads(data))
            for interface_mac, interface in device_list.items():
                for conn_type, conn_items in interface.items():
                    for dev_mac in conn_items:
                        mac = dev_mac.upper()
                        device = conn_items[mac]
                        ip = device.get("ip")
                        if not isinstance(ip, str):
                            ip = None
                        rssi = device.get("rssi", None)
                        if mac not in devices:
                            devices[mac] = _new_device(mac)
                        devices[mac].device_data["ip"] = ip
                        devices[mac].device_data["rssi"] = (
                            int(rssi) if rssi else None
                        )
                        devices[mac].interface["name"] = conn_type
                        devices[mac].interface["mac"] = interface_mac

        try:
            _parse_client_json(lines[0])
            _LOGGER.debug(
                "There are %s devices found after clientlist.json",
                len(devices),
            )
        except (TypeError, ValueError, AssertionError) as ex:
            _LOGGER.warning("Unable to parse clientlist.json")
            print(lines[0])
            print(ex)
        return devices

    async def get_connected_devices(
        self,
        reachable: bool = False,
    ) -> dict[str, Device]:
        """
        Retrieve devices and as much info as possible.

        If Settings.require_ip is `True,
            we filter out all devices that do not have ip mapped

        Calls various commands on the router and returns the superset of all
        responses. Some commands will not work on some routers.

        Args:
            reachable (bool): If true,
                filter out all devices that have not REACHABLE as status

        """
        devices: dict[str, Device] = await self._get_wl({})
        _LOGGER.debug(devices)
        devices = await self._get_arp(devices)
        _LOGGER.debug(devices)
        devices = await self._get_neigh(devices)
        _LOGGER.debug(devices)
        if self._settings.mode != Mode.AP:
            devices = await self._get_leases(devices)
            _LOGGER.debug(devices)

        devices = await self._filter_dev_list(devices)
        _LOGGER.debug(devices)
        if reachable:
            devices = {
                key: dev
                for key, dev in devices.items()
                if dev.device_data.get("status") == "REACHABLE"
            }
        if not self._settings.require_ip:
            return devices
        return {
            key: dev
            for key, dev in devices.items()
            if dev.device_data.get("ip") is not None
        }

    @property
    async def rx(self) -> int:
        """Get current RX total given in bytes."""
        return int(self._transfer_rates.rx)

    @property
    async def tx(self) -> int:
        """Get current RX total given in bytes."""
        return int(self._transfer_rates.tx)

    async def get_current_transfer_rates(
        self,
    ) -> tuple[int, int]:
        """Get current transfer rates calculated in per second in bytes."""
        _now = time()
        delay = _now - self._last_transfer_rates_check
        self._last_transfer_rates_check = _now
        eth0rx = eth0tx = 0
        vlanrx = vlantx = 0

        net_dev_lines = await self._connection.run_command(Command.NETDEV)
        if not net_dev_lines:
            _LOGGER.info("Unable to run %s", Command.NETDEV)
            return 0, 0

        for line in net_dev_lines[2:]:
            parts = split(r"[\s:]+", line.strip())
            # NOTES:
            #  * assuming eth0 always comes before vlan1 in dev file
            #  * counted bytes wrap around at 0xFFFFFFFF
            if parts[0] == "eth0":
                eth0rx = int(parts[1])  # received bytes
                eth0tx = int(parts[9])  # transmitted bytes
            elif parts[0] == "vlan1":
                vlanrx = int(parts[1])  # received bytes
                vlantx = int(parts[9])  # transmitted bytes

        def handle32bitwrap(v: int) -> int:
            return v if v > 0 else v + 0xFFFFFFFF

        # the true amount of Internet related data equals eth0 - vlan1
        inetrx = handle32bitwrap(eth0rx - vlanrx)
        inettx = handle32bitwrap(eth0tx - vlantx)

        rx = int(handle32bitwrap(inetrx - self._transfer_rates.rx) / delay)
        tx = int(handle32bitwrap(inettx - self._transfer_rates.tx) / delay)

        self._transfer_rates = TransferRates(inetrx, inettx)

        return rx, tx

    async def current_transfer_human_readable(
        self,
    ) -> tuple[str, str] | None:
        """Get current transfer rates in a human readable format."""
        rx, tx = await self.get_current_transfer_rates()

        if rx and rx > 0 and tx and tx > 0:
            return f"{convert_size(rx)}/s", f" {convert_size(tx)}/s"
        return "0/s", "0/s"

    async def get_loadavg(self) -> list[float]:
        """Get loadavg."""
        loadavg = list(
            map(
                float,
                (await self._connection.run_command(Command.LOADAVG))[0].split(
                    " "
                )[0:3],
            )
        )
        return loadavg

    async def add_dns_record(
        self, hostname: str, ipaddress: str
    ) -> list[str] | None:
        """
        Add record to /etc/hosts and HUP dnsmask to catch this record.

        Args:
            hostname (str): Hostname to add
            ipaddress (str): IP address to add
        """
        return list(
            await self._connection.run_command(
                Command.ADDHOST.format(hostname=hostname, ipaddress=ipaddress)
            )
        )

    async def get_interfaces_count(
        self,
    ) -> dict[str, dict[str, int]]:
        """Get counters for all network interfaces."""
        net_dev_lines = await self._connection.run_command(Command.NETDEV)
        lines = map(
            lambda i: list(filter(lambda j: j != "", i.split(" "))),
            net_dev_lines[2:-1],
        )
        interfaces: Iterable[tuple[str, dict[str, int]]] = map(
            lambda i: (
                i[0][0:-1],
                dict(zip(NETDEV_FIELDS, map(int, i[1:]))),
            ),
            lines,
        )
        return dict(interfaces)

    async def _find_temperature_commands(self) -> dict[str, float]:
        """Find which temperature commands work with the router, if any."""
        ret: dict[str, float] = {}

        for interface, commands in TEMP_COMMANDS.items():
            temp_command: TempCommand
            ret[interface] = 0.0
            for temp_command in commands:
                try:
                    result = await _run_temp_command(
                        self._connection, temp_command
                    )
                    if result:
                        self._temps_commands[interface] = temp_command
                        ret[interface] = result
                        break
                except (ValueError, IndexError, OSError):
                    continue
        return ret

    async def get_temperature(self) -> dict[str, float]:
        """Get temperature values we can find."""
        result: dict[str, float] = {}
        if not self._temps_commands:
            return await self._find_temperature_commands()
        for interface, command in self._temps_commands.items():
            temp = await _run_temp_command(self._connection, command)
            if temp:
                result[interface] = temp
        return result

    async def get_vpn_clients(self) -> list[dict[str, str]]:
        """Get current vpn clients."""
        data = await self.get_nvram("VPN")
        vpn_list = data["vpnc_clientlist"]

        vpns: list[dict[str, str]] = []
        for m in finditer(Regex.VPN_LIST, vpn_list):
            vpn_id = m.group("id")
            pid = await self._connection.run_command(
                Command.GET_PID_OF.format(name=f"vpnclient{vpn_id}")
            )

            vpn = {k: v for k, v in m.groupdict().items() if v}
            vpn_state_key = f"vpn_client{vpn_id}_state"
            vpn_state = int(data.get(vpn_state_key, 0))

            if vpn_state == 0 or not pid:
                vpn["state"] = "off"
            elif vpn_state == 1:
                vpn["state"] = "starting"
            elif vpn_state == 2:
                vpn["state"] = "on"

            vpns.append(vpn)

        return vpns

    async def start_vpn_client(self, vpn_id: int) -> list[str]:
        """
        Start a vpn client by id.

        Args:
            vpn_id (int): The id of the VPN service to start
        """
        for no in range(VPN_COUNT):
            _ = await self._connection.run_command(
                Command.VPN_STOP.format(id=no + 1)
            )

        return list(
            await self._connection.run_command(
                Command.VPN_START.format(id=vpn_id)
            )
        )

    async def stop_vpn_client(self, vpn_id: int) -> list[str]:
        """
        Stop a vpn client by id.

        Args:
            vpn_id (int): The id of the VPN service to stop
        """
        return list(
            await self._connection.run_command(
                Command.VPN_STOP.format(id=vpn_id)
            )
        )

    @property
    def is_connected(self) -> bool:
        """Is connected property."""
        return bool(self._connection.is_connected)

    async def disconnect(self) -> None:
        """Disconnect from router."""
        await self._connection.disconnect()
